function isOk = VGraph_AddInterclusterEdges(i, j)
global VGraph Obstacles

% Structure VGraph
%                 .ActiveIndices        -- A vector containing the indices of the active clusters ( = [1,2,ObstaclesInSight+2])
%                 .VertexCluster{i}     -- All the vertices associated with cluster i
%                                           ( i=1 -> start position - the only dynamic cluster
%                                             i=2 -> goal position
%                                             i=p -> Obstacle number p-2 )
%                                  .Vertices -- 3xV matrix. First 2 elements of each column represent coordinates of the vertices
%                                                           Last element indicating the validity of the vertices
%
%                 .InterClusterEdges{i,j}  -- Edges joining clusters i and j [IMPORTANT: i <= j ]
%                                        .isComputed -- Indicates whether these edges have been computed before (Check using isfield)
%                                        .Edges -- A 6xE matrix of edges along with cost
%                                                                   = [ v_i1    v_i2    ...   ;
%                                                                       v_j1    v_j2    ...   ;
%                                                                       cost_1  cost_2  ...   ;
%                                                                       a_1     a_2     ...   ;
%                                                                       b_1     b_2     ...   ;
%                                                                       c_1     c_2     ...   ; ]
%                                                                   v_kp is the index of the vertex in VGraph.VertexCluster(k).Vertices
%
%                 .LeastCostPath(k) -- contains information about the k-th segment of the path
%                                  .i    -- Index of cluster from where the edge originates
%                                  .v_i  -- Vertex index of cluster i
%                                  .j    -- Index of cluster where the edge ends
%                                  .v_j  -- Vertex index of cluster j
%                                  .e    -- Edge index in VGraph.InterClusterEdges{i,j}.Edges
%
% c_k = norm( VGraph.VertexCluster{i}.Vertices(1:2,v_ik) - VGraph.VertexCluster{j}.Vertices(1:2,v_jk) )
%

if i >= j
    isOk = 0;
    return;
end


isComputed = 0;
if all(size(VGraph.InterClusterEdges) >= [i,j])
    if isfield(VGraph.InterClusterEdges{i,j}, 'isComputed')
        isComputed = 1;
    end
end

% Adding tangents between clusters i and j (no checking done here)
if ~isComputed || size(VGraph.InterClusterEdges{i,j}.Edges,2)==0
    VGraph.InterClusterEdges{i,j}.Edges = [];
    allEdges = [];
    
    for vi = 1:size(VGraph.VertexCluster{i}.Vertices,2)
        for vj = 1:size(VGraph.VertexCluster{j}.Vertices,2)
            
            pi = VGraph.VertexCluster{i}.Vertices(:,vi);
            pj = VGraph.VertexCluster{j}.Vertices(:,vj);
            
            if ~(pi(3) & pj(3)) % Any one of the vertices is not valid!
                edgeValid = 0;
            else
                edgeValid = 1;
            end
            
            % Check if the edge made by pi-pj is a tangent to clusters i and j
            AuxiliaryColumn = ComputeAuxiliaryEdgeColumn(pi, pj);
            AuxiliaryColumn(1) = AuxiliaryColumn(1) + not(edgeValid)*1e6;
            
            points_i = VGraph.VertexCluster{i}.Vertices(1:2,setdiff(1:size(VGraph.VertexCluster{i}.Vertices,2),vi));
            points_j = VGraph.VertexCluster{j}.Vertices(1:2,setdiff(1:size(VGraph.VertexCluster{j}.Vertices,2),vj));

            vertexSigns_i = sign( AuxiliaryColumn(2)*points_i(1,:) + AuxiliaryColumn(3)*points_i(2,:) + AuxiliaryColumn(4) );
            vertexSigns_j = sign( AuxiliaryColumn(2)*points_j(1,:) + AuxiliaryColumn(3)*points_j(2,:) + AuxiliaryColumn(4) );
            if isempty(vertexSigns_i)
                SameSide_i = 1;
            else
                min_vertexSigns_i = min(vertexSigns_i);
                max_vertexSigns_i = max(vertexSigns_i);
                SameSide_i = (min_vertexSigns_i==0 | min_vertexSigns_i==1) | (max_vertexSigns_i==0 | max_vertexSigns_i==-1);
            end
            if isempty(vertexSigns_j)
                SameSide_j = 1;
            else
                min_vertexSigns_j = min(vertexSigns_j);
                max_vertexSigns_j = max(vertexSigns_j);
                SameSide_j = (min_vertexSigns_j==0 | min_vertexSigns_j==1) | (max_vertexSigns_j==0 | max_vertexSigns_j==-1);
            end

            if SameSide_i & SameSide_j % it's a tangent!
                VGraph.InterClusterEdges{i,j}.Edges = [VGraph.InterClusterEdges{i,j}.Edges, [vi;vj;AuxiliaryColumn]];
            end
            
            allEdges = [allEdges, [vi;vj;AuxiliaryColumn]];
            
        end
    end
    VGraph.InterClusterEdges{i,j}.isComputed = 1;
end


% If the point is inside an obstacle
if size(VGraph.InterClusterEdges{i,j}.Edges,2)==0 && (i==1 || j==1) && (i~=j)
    [SortedEdgeLength, SortIndex] = sort(allEdges(3,:));
    VGraph.InterClusterEdges{i,j}.Edges = allEdges(:,SortIndex(1:2));
end


% Check if the added edges are intersecting with any ActiveIndices clusters
checkClusters = setdiff(VGraph.ActiveIndices, [1 2 i j]);
goodEdges = [];
for e = 1:size(VGraph.InterClusterEdges{i,j}.Edges,2)
    EdgeDetails.i = i;
    EdgeDetails.v_i = VGraph.InterClusterEdges{i,j}.Edges(1,e);
    EdgeDetails.j = j;
    EdgeDetails.v_j = VGraph.InterClusterEdges{i,j}.Edges(2,e);
    EdgeDetails.e = e;
    
    isValid = 1;
    for ChkClusterNo = checkClusters
        if isIntersect_EdgeCluster(EdgeDetails, ChkClusterNo)
            isValid = 0;
            break;
        end
    end
    if isValid
        goodEdges = [goodEdges, e];
    end
    
end
VGraph.InterClusterEdges{i,j}.Edges = VGraph.InterClusterEdges{i,j}.Edges(:,goodEdges);

isOk = 1;
