function VGraph_AddObstacle(ObstacleNo)
global VGraph Obstacles

% Structure Obstacles
%                    .N                -- number of obstacles
%                    .obs{i}           -- cell array of structure that defines i-th obstacle
%                            .shape    -- Can be 'circle' or 'polygon' (present code supports only polygons)
%                       polygon:
%                            .Vertices -- 3xV matrix. First 2 elements of each column represent coordinates of the vertices
%                                                           Last element indicating the validity of the vertices
%

% Structure VGraph
%                 .ActiveIndices        -- A vector containing the indices of the active clusters ( = [1,2,ObstaclesInSight+2])
%                 .VertexCluster{i}     -- All the vertices associated with cluster i
%                                           ( i=1 -> start position - the only dynamic cluster
%                                             i=2 -> goal position
%                                             i=p -> Obstacle number p-2 )
%                                  .Vertices -- 3xV matrix. First 2 elements of each column represent coordinates of the vertices
%                                                           Last element indicating the validity of the vertices
%
%                 .InterClusterEdges{i,j}  -- Edges joining clusters i and j [IMPORTANT: i <= j ]
%                                        .isComputed -- Indicates whether these edges have been computed before (Check using isfield)
%                                        .Edges -- A 6xE matrix of edges along with cost
%                                                                   = [ v_i1    v_i2    ...   ;
%                                                                       v_j1    v_j2    ...   ;
%                                                                       cost_1  cost_2  ...   ;
%                                                                       a_1     a_2     ...   ;
%                                                                       b_1     b_2     ...   ;
%                                                                       c_1     c_2     ...   ; ]
%                                                                   v_kp is the index of the vertex in VGraph.VertexCluster(k).Vertices
%
%                 .LeastCostPath(k) -- contains information about the k-th segment of the path
%                                  .i    -- Index of cluster from where the edge originates
%                                  .v_i  -- Vertex index of cluster i
%                                  .j    -- Index of cluster where the edge ends
%                                  .v_j  -- Vertex index of cluster j
%                                  .e    -- Edge index in VGraph.InterClusterEdges{i,j}.Edges
%
% c_k = norm( VGraph.VertexCluster{i}.Vertices(1:2,v_ik) - VGraph.VertexCluster{j}.Vertices(1:2,v_jk) )
%

ClusterNo = ObstacleNo + 2;
VGraph.ActiveIndices = union(VGraph.ActiveIndices, ClusterNo);
VGraph.ActiveIndices = sort(VGraph.ActiveIndices);


% Adding the cluster
VGraph.VertexCluster{ClusterNo}.Vertices = Obstacles.obs{ObstacleNo}.Vertices;
VertexCountThisCluster = size(VGraph.VertexCluster{ClusterNo}.Vertices, 2);


% Adding self-edges
selfEdgesExist = 0;
if all(ClusterNo <= size(VGraph.VertexCluster))
    if isfield(VGraph.InterClusterEdges{ClusterNo,ClusterNo},'isComputed')
        selfEdgesExist = 1;
    end
end

if ~selfEdgesExist
    StartVertices = 1:VertexCountThisCluster;
    EndVertices = [2:VertexCountThisCluster, 1];
    EdgeValidity = VGraph.VertexCluster{ClusterNo}.Vertices(3,StartVertices) & VGraph.VertexCluster{ClusterNo}.Vertices(3,EndVertices);
    %StartVertices = StartVertices(EdgeValidity);
    %EndVertices = EndVertices(EdgeValidity);

    VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges = ...
        [ StartVertices ; EndVertices ;
        ComputeAuxiliaryEdgeColumn(VGraph.VertexCluster{ClusterNo}.Vertices(1:2,StartVertices), VGraph.VertexCluster{ClusterNo}.Vertices(1:2,EndVertices)) ];
    VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges(3,:) = VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges(3,:) + not(EdgeValidity)*1e6;
    
    VGraph.InterClusterEdges{ClusterNo,ClusterNo}.isComputed = 1;
end


% Add the edges connecting to other active clusters
% Checking for intersection with these edges with other active polygons.
for k = setdiff(VGraph.ActiveIndices, ClusterNo)
    
    i = min([k, ClusterNo]);
    j = max([k, ClusterNo]);
    
    isOk = VGraph_AddInterclusterEdges(i, j);
    if ~isOk
        disp('WARNING: something wrong!');
    end
    
end


% Check for intersection of this new obstacle with edges running between other existing clusters
VGraph.ActiveIndices = sort(VGraph.ActiveIndices);
checkClusters = setdiff(VGraph.ActiveIndices, ClusterNo);
for a = 1:length(checkClusters)
    for b = (a+1):length(checkClusters)
        i = checkClusters(a);
        j = checkClusters(b);
        
        goodEdges = [];
        for e = 1:size(VGraph.InterClusterEdges{i,j}.Edges,2)
            EdgeDetails.i = i;
            EdgeDetails.v_i = VGraph.InterClusterEdges{i,j}.Edges(1,e);
            EdgeDetails.j = j;
            EdgeDetails.v_j = VGraph.InterClusterEdges{i,j}.Edges(2,e);
            EdgeDetails.e = e;
            if ~isIntersect_EdgeCluster(EdgeDetails, ClusterNo)
                goodEdges = [goodEdges, e];
            end
        end
        VGraph.InterClusterEdges{i,j}.Edges = VGraph.InterClusterEdges{i,j}.Edges(:,goodEdges);
        
    end
end

