function result = isIntersect_EdgeObstacle(EdgeDetails, ObstacleNo)
global VGraph Obstacles

% Structure EdgeDetails
%                      .i    -- Index of cluster from where the edge originates
%                      .v_i  -- Vertex index of cluster i
%                      .j    -- Index of cluster where the edge ends
%                      .v_j  -- Vertex index of cluster j
%                      .e    -- Edge index in VGraph.InterClusterEdges{i,j}.Edges
%

% Structure Obstacles
%                    .N                -- number of obstacles
%                    .obs{i}           -- cell array of structure that defines i-th obstacle
%                            .shape    -- Can be 'circle' or 'polygon' (present code supports only polygons)
%                       polygon:
%                            .Vertices -- 3xV matrix. First 2 elements of each column represent coordinates of the vertices
%                                                           Last element indicating the validity of the vertices
%

% Structure VGraph
%                 .ActiveIndices        -- A vector containing the indices of the active clusters ( = [1,2,ObstaclesInSight+2])
%                 .VertexCluster{i}     -- All the vertices associated with cluster i
%                                           ( i=1 -> start position - the only dynamic cluster
%                                             i=2 -> goal position
%                                             i=p -> Obstacle number p-2 )
%                                  .Vertices -- 3xV matrix. First 2 elements of each column represent coordinates of the vertices
%                                                           Last element indicating the validity of the vertices
%
%                 .InterClusterEdges{i,j}  -- Edges joining clusters i and j [IMPORTANT: i <= j ]
%                                        .isComputed -- Indicates whether these edges have been computed before (Check using isfield)
%                                        .Edges -- A 6xE matrix of edges along with cost
%                                                                   = [ v_i1    v_i2    ...   ;
%                                                                       v_j1    v_j2    ...   ;
%                                                                       cost_1  cost_2  ...   ;
%                                                                       a_1     a_2     ...   ;
%                                                                       b_1     b_2     ...   ;
%                                                                       c_1     c_2     ...   ; ]
%                                                                   v_kp is the index of the vertex in VGraph.VertexCluster(k).Vertices
%
%                 .LeastCostPath(k) -- contains information about the k-th segment of the path
%                                  .i    -- Index of cluster from where the edge originates
%                                  .v_i  -- Vertex index of cluster i
%                                  .j    -- Index of cluster where the edge ends
%                                  .v_j  -- Vertex index of cluster j
%                                  .e    -- Edge index in VGraph.InterClusterEdges{i,j}.Edges
%
% c_k = norm( VGraph.VertexCluster{i}.Vertices(1:2,v_ik) - VGraph.VertexCluster{j}.Vertices(1:2,v_jk) )
%

ClusterNo = ObstacleNo + 2;

if all(ClusterNo <= size(VGraph.VertexCluster))
    if isfield(VGraph.InterClusterEdges{ClusterNo,ClusterNo},'isComputed')
        result = isIntersect_EdgeCluster(EdgeDetails, ClusterNo);
        return;
    end
end

% Create the self-edges for the cluster
VGraph.VertexCluster{ClusterNo}.Vertices = Obstacles.obs{ObstacleNo}.Vertices;
VertexCountThisCluster = size(VGraph.VertexCluster{ClusterNo}.Vertices, 2);

StartVertices = 1:VertexCountThisCluster;
EndVertices = [2:VertexCountThisCluster, 1];
EdgeValidity = VGraph.VertexCluster{ClusterNo}.Vertices(3,StartVertices) & VGraph.VertexCluster{ClusterNo}.Vertices(3,EndVertices);
% StartVertices = StartVertices(EdgeValidity);
% EndVertices = EndVertices(EdgeValidity);

VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges = ...
        [ StartVertices ; EndVertices ;
          ComputeAuxiliaryEdgeColumn(VGraph.VertexCluster{ClusterNo}.Vertices(1:2,StartVertices), VGraph.VertexCluster{ClusterNo}.Vertices(1:2,EndVertices)) ];
VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges(3,:) = VGraph.InterClusterEdges{ClusterNo,ClusterNo}.Edges(3,:) + not(EdgeValidity)*1e6;

VGraph.InterClusterEdges{ClusterNo,ClusterNo}.isComputed = 1;

result = isIntersect_EdgeCluster(EdgeDetails, ClusterNo);
