/*
 *  Gazebo - Outdoor Multi-Robot Simulator
 *  Copyright (C) 2003  
 *     Nate Koenig & Andrew Howard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <gazebo.h>
#include <gazebo/World.hh>
#include <gazebo/WorldFile.hh>
#include <gazebo/ModelFactory.hh>
#include <gazebo/Body.hh>
#include <gazebo/BoxGeom.hh>
#include <gazebo/SphereGeom.hh>
#include <gazebo/WheelGeom.hh>
#include <gazebo/HingeJoint.hh>
#include <gazebo/BallJoint.hh>

#include <ctime>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "Scarab_Noisy.hh"

#define BODY_WIDTH 0.135
#define BODY_LENGTH 0.199868
#define BODY_HEIGHT 0.206732
#define BODY_THICKNESS 0.005
#define TOTAL_HEIGHT 0.221732
#define WHEEL_SEP 0.210
#define WHEEL_DIAM 0.100
#define WHEEL_THICK 0.02392
#define CASTER_OFFSET_X 0.029125
#define CASTER_OFFSET_Y 0.0941
#define CASTER_HEIGHT 0.03119
#define BODY_Z -(BODY_HEIGHT/2 + (TOTAL_HEIGHT - BODY_HEIGHT))

#define WHEEL_R 0.3
#define WHEEL_G 0.3
#define WHEEL_B 0.3
#define WHEEL_A 0.5

#define BODY_R 0.8
#define BODY_G 0.4
#define BODY_B 0.1
#define BODY_A 1.0

#define CASTER_R 0.8
#define CASTER_G 0.8
#define CASTER_B 0.8
#define CASTER_A 0.5

/////////////////////////////////////////////////////////////////////////////
// Register the model
GZ_REGISTER_PLUGIN("Scarab_Noisy", Scarab)

//////////////////////////////////////////////////////////////////////////////
// Constructor
Scarab::Scarab( World *world )
    : Model( world )
{
	int i;
	this->chassis = NULL;
	for(i=0; i < 2; i++) {
		this->wheels[i] = NULL;
		this->wheelJoints[i] = NULL;
                this->casterWheels[i] = NULL;
                this->casterJoints[i] = NULL;
	}

	r = gsl_rng_alloc (gsl_rng_taus);
	gsl_rng_set(r, time(0));
        
  return;
}

//////////////////////////////////////////////////////////////////////////////
// Destructor
Scarab::~Scarab()
{

	int i;
	if (this->chassis)
		delete this->chassis;

	this->chassis = NULL;
	for(i=0; i<2; ++i) {
		if(this->wheels[i])
                  delete this->wheels[i];

		if(this->wheelJoints[i])
                  delete this->wheelJoints[i];

                if(this->casterWheels[i])
                  delete this->casterWheels[i];

                if(this->casterJoints[i])
                  delete this->casterJoints[i];

		this->wheels[i] = NULL;
		this->wheelJoints[i] = NULL;
                this->casterWheels[i] = NULL;
                this->casterJoints[i] = NULL;
	}

	gsl_rng_free(r);

  return;
}


//////////////////////////////////////////////////////////////////////////////
// Load the sensor
int Scarab::Load( WorldFile *file, WorldFileNode *node )
{

  this->wheelSep = WHEEL_SEP;
  this->wheelDiam = WHEEL_DIAM;

  this->updatePeriod = 1.0 / (node->GetDouble("updateRate", 10) + 1e-6);
  this->updateTime = -updatePeriod;

  // Create the ODE objects
  if (this->OdeLoad(file, node) != 0)
    return -1;

  // Initial battery level
  this->batteryLevel = node->GetDouble("batteryLevel", 12.4);

  // Discharge curve: about one hour quiescent
  this->batteryCurve[0] = node->GetTupleDouble("batteryCurve", 0, 2 / 3600.0);
  this->batteryCurve[1] = node->GetTupleDouble("batteryCurve", 1, 2 / 1e4);  

	this->cov_v = node->GetTupleDouble("output_noise", 0, 0.0);
	this->cov_w = node->GetTupleDouble("output_noise", 1, 0.0);
	this->cov_vw = node->GetTupleDouble("output_noise", 2, 0.0);

	this->input_cov_v = node->GetTupleDouble("input_noise", 0, 0.0);
	this->input_cov_w = node->GetTupleDouble("input_noise", 1, 0.0);
	this->input_cov_vw = node->GetTupleDouble("input_noise", 2, 0.0);

  return 0;
}

int Scarab::OdeLoad(WorldFile *file, WorldFileNode *node)
{
  int i;
  Geom *geom;
  GzVector a;

  // Note: model origin is the center of the drive axle.  
  double mass = 5;
  double casterMass = 3;

  GzVector bodySize = GzVectorSet(BODY_LENGTH, BODY_WIDTH, BODY_HEIGHT);
  GzVector bodyPos = GzVectorSet(0, 0, BODY_Z);

  GzVector wheelSize = GzVectorSet(this->wheelDiam, this->wheelDiam, WHEEL_THICK);
  GzVector wheelPos = GzVectorSet(0, this->wheelSep/2, bodyPos.z - (TOTAL_HEIGHT - 
                                                                    (BODY_HEIGHT + WHEEL_DIAM)/2));

  GzVector casterSize = GzVectorSet(CASTER_HEIGHT, CASTER_HEIGHT, CASTER_HEIGHT);
  GzVector casterPos1 = GzVectorSet(bodyPos.x + BODY_LENGTH/2 - CASTER_OFFSET_X, CASTER_OFFSET_Y - BODY_WIDTH/2,
                                    wheelPos.z - wheelSize.x/2 + casterSize.x/2);


  GzVector casterPos2 = GzVectorSet(bodyPos.x - BODY_LENGTH/2 + CASTER_OFFSET_X, -(CASTER_OFFSET_Y - BODY_WIDTH/2),
                                    wheelPos.z - wheelSize.x/2 + casterSize.x/2);

  // Set the body and wheel color
  this->body_color = node->GetColor("bodyColor", 
                                    GzColor(BODY_R, BODY_G, BODY_B, BODY_A));
  this->wheel_color = node->GetColor("wheelColor", 
                                     GzColor(WHEEL_R, WHEEL_G, WHEEL_B, WHEEL_A));
  
  // Create the main chassis of the robot
  this->chassis = new Body( this->world );

  // body

  geom = new BoxGeom(this->chassis, this->modelSpaceId, 
                     bodySize.x - 2*BODY_THICKNESS, 
                     bodySize.y - 2*BODY_THICKNESS, 
                     bodySize.z - 2*BODY_THICKNESS);
  geom->SetRelativePosition( GzVectorSet(bodyPos.x, bodyPos.y, bodyPos.z) );
  geom->SetMass( 0.0 );
  geom->SetColor(this->body_color);  // Changed

  geom = new BoxGeom(this->chassis, this->modelSpaceId, bodySize.x, bodySize.y, bodySize.z);
  geom->SetRelativePosition( GzVectorSet(bodyPos.x, bodyPos.y, bodyPos.z) );
  geom->SetMass( mass );
  geom->SetColor(this->body_color);


  // wheel blocks
  geom = new BoxGeom(this->chassis, this->modelSpaceId, this->wheelDiam*0.7, 
                     (WHEEL_SEP/2)-(WHEEL_THICK)-(bodySize.y/2), this->wheelDiam*0.5);
  geom->SetRelativePosition( GzVectorSet(0, (wheelPos.y - WHEEL_THICK*1.5), wheelPos.z) );
  geom->SetMass( 0 );
  geom->SetColor(this->body_color);

  geom = new BoxGeom(this->chassis, this->modelSpaceId, this->wheelDiam*0.7, 
                     (WHEEL_SEP/2)-(WHEEL_THICK)-(bodySize.y/2), this->wheelDiam*0.5);
  geom->SetRelativePosition( GzVectorSet(0, -(wheelPos.y - WHEEL_THICK*1.5), wheelPos.z) );
  geom->SetMass( 0 );
  geom->SetColor(this->body_color);

  this->AddBody( this->chassis, true );

  uint wheelCategory = (uint)node->GetInt("wheelCategory", 2);
  uint bumperCategory = (uint)node->GetInt("bumperCategory", 1);

  // Create the wheels
  for (i = 0; i < 2; i++)
  {
    this->wheels[i] = new Body( this->world );

    geom = new WheelGeom(this->wheels[i], this->modelSpaceId, 0.5 * wheelSize.x, wheelSize.z);
    geom->SetRelativePosition( GzVectorSet(0, 0, 0) );
    geom->SetMass( 0.5 );
    geom->SetColor(this->wheel_color);
    geom->SetCategoryBits(wheelCategory);
    geom->SetCollideBits(~bumperCategory);
    this->AddBody( this->wheels[i] );
  }

  // Set the wheel positions
  this->wheels[0]->SetPosition(GzVectorSet(wheelPos.x, +wheelPos.y, wheelPos.z));
  this->wheels[0]->SetRotation(GzQuaternFromAxis(1, 0, 0, -M_PI / 2));
  this->wheels[1]->SetPosition(GzVectorSet(wheelPos.x, -wheelPos.y, wheelPos.z));
  this->wheels[1]->SetRotation(GzQuaternFromAxis(1, 0, 0, +M_PI / 2));

  // Attach the wheels to the chassis at their current positions
  for (i = 0; i < 2; i++)
  {
    this->wheelJoints[i] = new HingeJoint( this->world );
    this->wheelJoints[i]->Attach( this->wheels[i], this->chassis );
    a = wheels[i]->GetPosition();
    this->wheelJoints[i]->SetAnchor( a );
    this->wheelJoints[i]->SetAxis( GzVectorSet(0, 1, 0) );
    this->wheelJoints[i]->SetParam( dParamSuspensionERP, 0.4 );
    this->wheelJoints[i]->SetParam( dParamSuspensionCFM, 0.8 );
  }

  // Create the caster wheels
  for (i = 0; i < 2; i++)
  {
    this->casterWheels[i] = new Body(this->world);

    geom = new SphereGeom(this->casterWheels[i], this->modelSpaceId, casterSize.x/2);
    geom->SetRelativePosition( GzVectorSet(0, 0, 0) );
    geom->SetMass(casterMass);
    geom->SetColor(GzColor(CASTER_R, CASTER_G, CASTER_B, CASTER_A)); // grey-silver color

    this->AddBody(this->casterWheels[i]);
  }
  
  this->casterWheels[0]->SetPosition(casterPos1);
  this->casterWheels[0]->SetRotation(GzQuaternFromEuler(M_PI / 2, 0, 0));
  this->casterWheels[1]->SetPosition(casterPos2);
  this->casterWheels[1]->SetRotation(GzQuaternFromEuler(M_PI / 2, 0, 0)); 

  // Attach the caster to the chassis
  for (i = 0; i < 2; i++)
  {
    this->casterJoints[i] = new BallJoint(this->world);
    this->casterJoints[i]->Attach(this->casterWheels[i], this->chassis);
    a = this->casterWheels[i]->GetPosition();
    this->casterJoints[i]->SetAnchor( a );
  }
    
  return 0;

}

//////////////////////////////////////////////////////////////////////////////
// Initialize the sensor
int Scarab::Init( WorldFile *file, WorldFileNode *node )
{

	  // Create position interface
  this->position_iface = gz_position_alloc();
  assert(this->position_iface);
  if (gz_position_create(this->position_iface, this->world->gz_server, this->GetId(),
                         "Scarab", this->GetIntId(), this->GetParentIntId()) != 0)
    return -1;

  // Create power inteface
  this->power_iface = gz_power_alloc();
  if (gz_power_create(this->power_iface, this->world->gz_server, this->GetId(),
                      "Scarab", this->GetIntId(), this->GetParentIntId()) != 0)
    return -1;

  // Reset odometric pose
  this->odomPose[0] = 0.0;
  this->odomPose[1] = 0.0;
  this->odomPose[2] = 0.0;

  return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Finalize the sensor
int Scarab::Fini()
{

  // Close power interface
  gz_power_destroy( this->power_iface );
  gz_power_free( this->power_iface );
  this->power_iface = NULL;

  // Close position interface
  gz_position_destroy( this->position_iface );
  gz_position_free( this->position_iface );
  this->position_iface = NULL;

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Update the sensor information; returns non-zero if sensor data has
// been updated
void Scarab::Update( double step )
{

  // Do nothing if paused
  if (step == 0)
    return;

  // Update the odometry (do this always for better accuracy)
  this->UpdateOdometry( step );

  // Otherwise, update periodically
  if (this->world->GetSimTime() - this->updateTime < this->updatePeriod)
    return;
  this->updateTime = this->world->GetSimTime();

  // Get commands from the external interface
  this->GetPositionCmd();

  if (this->enableMotors)
  {
    this->wheelJoints[0]->SetParam( dParamVel, 
                                    this->wheelSpeed[1] / this->wheelDiam * 2 );
    this->wheelJoints[1]->SetParam( dParamVel, 
                                    this->wheelSpeed[0] / this->wheelDiam * 2 );
    

    /*this->wheelJoints[0]->SetTorque(this->wheelSpeed[0]);
    this->wheelJoints[1]->SetTorque(this->wheelSpeed[1]);
    */

    this->wheelJoints[0]->SetParam( dParamFMax, 1.1 );
    this->wheelJoints[1]->SetParam( dParamFMax, 1.1 );
  }
  else
  {
    this->wheelJoints[0]->SetParam( dParamVel, 0.0 ); 
    this->wheelJoints[1]->SetParam( dParamVel, 0.0 );
  }

  // Update the interfaces
  this->PutPositionData();
  this->PutPowerData();

  return;

}


//////////////////////////////////////////////////////////////////////////////
// Update the odometry
void Scarab::UpdateOdometry( double step )
{
  double wd, ws;
  double d1, d2;
  double dr, da;

  wd = this->wheelDiam;
  ws = this->wheelSep;

  // Average distance travelled by left and right wheels
  d1 = step * wd / 2 * 
    (wheelJoints[0]->GetAngleRate());
  d2 = step * wd / 2 * 
    (wheelJoints[1]->GetAngleRate());;

  dr = (d1 + d2) / 2;
  da = (d2 - d1) / ws;

	// Noise is drawn from gaussian and is proportional to magnitude
	// of input.
	double error_v = dr*gsl_ran_gaussian(r, cov_v) + da*gsl_ran_gaussian(r, cov_vw);
	double error_w = da*gsl_ran_gaussian(r, cov_w) + dr*gsl_ran_gaussian(r, cov_vw);
	dr += error_v;
	da += error_w;

	//printf("Output error: %f, %f\n", error_v, error_w);

  // Compute odometric pose
  this->odomPose[0] += dr * cos( this->odomPose[2] );
  this->odomPose[1] += dr * sin( this->odomPose[2] );
  this->odomPose[2] += da;

  // Compute odometric instantaneous velocity
  this->odomVel[0] = dr / step;
  this->odomVel[1] = 0.0;
  this->odomVel[2] = da / step;

  // Update the power discharge; this is probably completely bogus
  this->batteryLevel -= this->batteryCurve[0] * step;
  this->batteryLevel -= this->batteryCurve[1] * d1 * d1;
  this->batteryLevel -= this->batteryCurve[1] * d2 * d2;

  return;
}


//////////////////////////////////////////////////////////////////////////////
// Get commands from the external interface
void Scarab::GetPositionCmd()
{
  double vr, va;

  vr = this->position_iface->data->cmd_vel_pos[0];
  va = this->position_iface->data->cmd_vel_rot[2];

	// Noise is drawn from gaussian and is proportional to magnitude
	// of input. (assume cross terms are symmetric)
	double error_v = vr*gsl_ran_gaussian(r, input_cov_v) + va*gsl_ran_gaussian(r, input_cov_vw);
	double error_w = va*gsl_ran_gaussian(r, input_cov_w) + vr*gsl_ran_gaussian(r, input_cov_vw);
	vr += error_v;
	va += error_w;

	//printf("Input error: %f, %f\n", error_v, error_w);

  this->enableMotors = this->position_iface->data->cmd_enable_motors;
  this->wheelSpeed[0] = vr + va * this->wheelSep / 2;
  this->wheelSpeed[1] = vr - va * this->wheelSep / 2;
  // TESTING
  //this->wheelSpeed[0] = vr;
  //this->wheelSpeed[1] = va;
  
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the data in the erinterface
void Scarab::PutPositionData()
{

  gz_position_lock(this->position_iface, 1);
    
  // Data timestamp
  this->position_iface->data->time = this->world->GetSimTime();

  this->position_iface->data->pos[0] = this->odomPose[0];
  this->position_iface->data->pos[1] = this->odomPose[1];
  this->position_iface->data->rot[2] = this->odomPose[2];

  this->position_iface->data->vel_pos[0] = this->odomVel[0];
  this->position_iface->data->vel_rot[2] = this->odomVel[2];

  gz_position_unlock(this->position_iface);
  
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the data in the power interface
void Scarab::PutPowerData()
{
  gz_power_lock(this->power_iface, 1);
    
  this->power_iface->data->time = this->world->GetSimTime();
  this->power_iface->data->levels[0] = this->batteryLevel;

  gz_power_unlock(this->power_iface);
    
  return;
}

